//Camera Tracking Tetris: code by Aaron Steed 2006 - http://robotacid.com
//Game concept by John Holder and Aaron Steed 2006
import processing.video.*;
import processing.opengl.*;

//following variables are for setting up testing and for moving the code from computer to computer

final static boolean exhibitMode = true;              //true: any key drops blocks and move from title screen, camera capture is on
                                                       //      microphone is used to rotate blocks
                                                       //false:mouse impersonates capture signal, arrow keys UP rotates and DOWN drops blocks
                                                       //      although microphone can be used to rotate blocks as well. The ENTER key starts
                                                       //      the game and moves from the game over screen

//alter the following string to the capture device you are using, see the Processing video library reference to find out what you are using
//final static String captureDeviceName = "o100vc.dll - Osprey Capture Card 1-WDM";
final static String captureDeviceName = "Creative WebCam Live! Pro-WDM";

static boolean showCapture = false;                    //show the capture image and capture column representation
                                                       //set this to true to calibrate the effectiveness of tracking white

TetrisField field;                                     //Tetris game field
TetrisBlock block;                                     //active Tetris block
TetrisNextBlock nextBlock;                             //next Tetris block
Building building;                                     //building made from lines cleared
Capture capture;                                       //webcam capture object
final static int dropRate = 800;                       //initial rate at which to drop blocks in milliseconds
final static int fieldWide = 10;                       //width of Tetris playing field
final static int fieldHigh = 20;                       //height of Tetris playing field
final static int captureWide = 640;                    //width of capture
final static int captureHigh = 480;                    //height of capture
static int trackColumnWidth = captureWide / fieldWide; //width of columns for white tracking function
static int trackColumnWidthRect = 1024 / fieldWide;    //capture image tracking rectangle width
final static int trackSpeed = 4;                       //setting for skipping pixels in the white tracking function
static int locale = -1;                                //variable to hold white tracking report
static PImage [] architecture;                         //selection of architectural squares for use as textures
static PImage [][] mugShot;                            //selection of mug shots
final static int imagesToLoad = 9;                     //quantity of images to load
final static int mugsToLoad = 7;                       //ranks of mug shots to load
final static int [] mugShotTable = {                   //quantities of mug shots in each rank
  4, 1, 4, 1, 1, 1, 1};
static color [] blockColor;                            //colour table for Tetris square elements
static color [] blockColorEdge;                        //colour table for isometric edges of square elements
final static int squareFadeRate = 30;                  //frames of animation for blocks fading into field elements
final static int blockDepth = 20;                      //isometric depth of all blocks in game
final static int increaseSpeed = 3;                    //increase speed every "increaseSpeed" lines
static float rotateThreshold = 0.04;                   //microphone listening level for rotation
static boolean showMicrophone = true;                  //show graphic of microphone levels
static boolean microphoneLock = false;                 //lockout input from microphone till volume drops below threshold
static int gameState = 0;                              //status of game (title screen, game playing, game over)
final static int title = 0;                            //name of game state 0
final static int playGame = 1;                         //name of game state 1
final static int gameOver = 2;                         //name of game state 2
static int bestScore = 0;                              //high score variable
static int ranking = 0;                                //representation of player's skill based on score
static int rankingMugShot = 0;                         //the mug shot chosen at random for the current rank
static int lastScore = 0;                              //placeholder for score variable in TetrisField
static int lastLines = 0;                              //placeholder for lines variable in TetrisField
static int lastRanking = 0;                            //placeholder for rank
static final int [] rankTable = {                      //levels of score needed to attain ranks
  //0, 10, 20, 30, 40, 50, 60};
  0, 50, 100, 150, 200, 250, 300};
static final String [] rankName = {                    //titles of each rank
  "technical demonstrator", "technical manager", "senior lecturer", "subject director", "professor", "head of school", "anan"};

//quadText settings and variables
QuadTextWobble titleText;                              //
QuadTextWobble highScoreText;                          //objects that maintain the wobbly text in the game
QuadTextWobble scoreText;                              //
QuadTextWobble linesText;
QuadTextWobble nextPieceText;
QuadTextWobble rankText;
QuadTextWobble youScoredText;
QuadTextWobble andMadeText;
QuadTextWobble yourRankText;
final color textStroke = color(255);                   //colour of text
static final float wobbleRange = 0.3;                  //extent of wobble
static final float wobbleSpeed = 0.05;                 //speed of text wobble
static Character and;                                  //used to calculate unicode values of text
static final int [][] charTable = {                    //pathways of 3D isometric text, unicode indexed
  {0, 0, 2, 0, 2, 3, 0, 3, 0, 0},                      // 0
  {0, 1, 1, 0, 1, 3},                                  // 1
  {0, 0, 2, 0, 2, 1, 0, 3, 2, 3},                      // 2
  {0, 0, 2, 0, 0, 1, 2, 3, 0, 3},                      // 3
  {2, 2, 0, 2, 1, 0, 1, 3},                            // 4
  {2, 0, 0, 0, 0, 1, 2, 2, 2, 3, 0, 3},                // 5
  {2, 0, 0, 0, 0, 3, 2, 3, 2, 1, 0, 1},                // 6
  {0, 0, 2, 0, 0, 3},                                  // 7
  {0, 0, 2, 0, 0, 3, 2, 3, 0, 0},                      // 8
  {0, 3, 2, 3, 2, 0, 0, 0, 0, 2, 2, 2},                // 9
  {2, 1, 0, 3, 2, 3, 2, 1},                            // a (10)
  {0, 0, 0, 3, 2, 3, 2, 1, 0, 1},                      // b (11)
  {2, 1, 0, 1, 0, 3, 2, 3},                            // c (12)
  {2, 0, 2, 3, 0, 3, 0, 1, 2, 1},                      // d (13)
  {0, 2, 2, 2, 2, 1, 0, 1, 0, 3, 2, 3},                // e (14)
  {0, 3, 0, 0, 2, 0, 0, 1, 2, 1},                      // f (15)
  {2, 3, 0, 3, 0, 1, 2, 1, 2, 4, 0, 4},                // g (16)
  {0, 0, 0, 3, 1, 1, 2, 1, 2, 3},                      // h (17)
  {1, 1, 1, 3},                                        // i (18)
  {1, 0, 1, 4, 0, 4},                                  // j (19)
  {0, 0, 0, 3, 2, 3, 0, 2, 2, 1},                      // k (20)
  {1, 0, 1, 3},                                        // l (21)
  {0, 3, 0, 1, 1, 3, 2, 1, 2, 3},                      // m (22)
  {0, 3, 0, 1, 2, 1, 2, 3},                            // n (23)
  {0, 1, 2, 1, 2, 3, 0, 3, 0, 1},                      // o (24)
  {0, 4, 0, 1, 2, 1, 2, 3, 0, 3},                      // p (25)
  {2, 4, 2, 1, 0, 1, 0, 3, 2, 3},                      // q (26)
  {0, 3, 0, 1, 2, 1},                                  // r (27)
  {2, 1, 0, 1, 0, 2, 2, 2, 2, 3, 0, 3},                // s (28)
  {1, 3, 1, 0, 0, 1, 2, 1},                            // t (29)
  {0, 1, 0, 3, 2, 3, 2, 1},                            // u (30)
  {0, 1, 1, 3, 2, 1},                                  // v (31)
  {0, 1, 0, 3, 1, 1, 2, 3, 2, 1},                      // w (32)
  {0, 1, 2, 3, 0, 3, 2, 1},                            // x (33)
  {0, 1, 0, 3, 2, 1, 2, 4, 0, 4},                      // y (34)
  {0, 1, 2, 1, 0, 3, 2, 3}                             // z (35)
  };
QuadChar [] alphabet = new QuadChar[36];               //array of 3D isometric characters
static final int titleSize = 16;                       //size of text during title state
static final int playGameSize = 6;                     //size of text during playGame state
static final int gameOverSize = 10;                    //size of text during gameOver state
