//Authors note on re-using this code:
//
//The field, block and nextBlock objects are interwoven and singular in nature
//They cannot be cast again and again as they refer to each other's properties
//They have been structured as classes for easier reading and adminstration
//TetrisField is cast as "field", Tetris block is cast as "block", TetrisNextBlock is cast as nextBlock

//The Tetris game is an array of square elements that fill up slowly by the
//decending TetrisBlock object which is converted into elements when it contacts
//square elements below it. Upon conversion the array is checked for a width
//full line, that line is then cleared, score is increased and the blocks above
//are scrolled down. To make collision detection a lot simpler, the field is bounded
//by two lines of blocks that are ignored during the game but can be seen by setting
//"drawBorder" to true
class TetrisField{
  int x, y, depth, wide, realWide, high, sizex, sizey, lines, timer, dropRate, increaseSpeed, score;
  int [] square;
  int [] squareFade;
  boolean drawBorder = false;
  TetrisField(int x, int y, int wide, int high, int sizex, int sizey, int dropRate, int increaseSpeed){
    this.x = x;                                      //drawing handle
    this.y = y;                                      //drawing handle
    depth = blockDepth;                              //isometric depth of square elements
    this.wide = wide;                                //columns in visible field
    this.realWide = wide + 4;                        //columns including border
    this.high = high;                                //vertical rows in field
    this.sizex = sizex;                              //width of square elements
    this.sizey = sizey;                              //height of square elements
    lines = 0;                                       //lines created by player
    timer = millis();                                //timer for dropping the block
    this.dropRate = dropRate;                        //time in milliseconds before block is dropped a vertical row
    this.increaseSpeed = increaseSpeed;              //increase speed every "increaseSpeed" lines
    score = 0;                                       //score awarded for playing well
    square = new int[realWide * (high + 2)];         //square elements array
    squareFade = new int[square.length];             //timer for apparent fading blocks into elements
    //establishing of border
    for(int i = 0; i < realWide; i++){
      square[i + high * realWide] = 1;
      square[i + (high + 1) * realWide] = 1;
    }
    for(int i = 0; i < high; i++){
      square[i * realWide] = 1;
      square[1 + i * realWide] = 1;
      square[realWide - 1 + (i * realWide)] = 1;
      square[realWide - 2 + (i * realWide)] = 1;
    }
  }
  void draw(){
    stroke(255);
    noFill();
    rect(x, y, wide * sizex, high * sizey);
    beginShape();
    vertex(x - (depth>>1), y - (depth>>1), -depth);
    vertex(x + (wide * sizex) - (depth>>1), y - (depth>>1), -depth);
    vertex(x + (wide * sizex) - (depth>>1), y + (high * sizey) - (depth>>1), -depth);
    vertex(x - (depth>>1), y + (high * sizey) - (depth>>1), -depth);
    endShape();
    for(int i = 0; i < wide; i++){
      for(int j = 0; j < high; j++){
        if(square[i + 2 + j * realWide] > 0){
          int xDraw = x + i * sizex;
          int yDraw = y + j * sizey;
          noStroke();
          fill(blockColorEdge[squareFade[i + 2 + j * realWide]]);
          architecture[square[i + 2 + j * realWide]].updatePixels();
          beginShape(QUADS);
          texture(architecture[square[i + 2 + j * realWide]]);
          vertex(xDraw - (depth>>1), yDraw - (depth>>1), -depth, 0, 0);
          vertex(xDraw, yDraw, 0, sizex, 0);
          vertex(xDraw, yDraw + sizey, 0, sizex, sizey);
          vertex(xDraw - (depth>>1), yDraw + sizey - (depth>>1), -depth, 0, sizey);
          endShape();
          fill(blockColorEdge[squareFade[i + 2 + j * realWide]]);
          beginShape(QUADS);
          texture(architecture[square[i + 2 + j * realWide]]);
          vertex(xDraw - (depth>>1), yDraw - (depth>>1), -depth, 0, 0);
          vertex(xDraw + sizex - (depth>>1), yDraw - (depth>>1), -depth, sizex, 0);
          vertex(xDraw + sizex, yDraw, 0, sizex, sizey);
          vertex(xDraw, yDraw, 0, 0, sizey);
          endShape();
          fill(blockColor[squareFade[i + 2 + j * realWide]]);
          beginShape(QUADS);
          texture(architecture[square[i + 2 + j * realWide]]);
          vertex(xDraw, yDraw, 0, 0, 0);
          vertex(xDraw + sizex, yDraw, 0, sizex, 0);
          vertex(xDraw + sizex, yDraw + sizey, 0, sizex, sizey);
          vertex(xDraw, yDraw + sizey, 0, 0, sizey);
          endShape();
          if(squareFade[i + 2 + j * realWide] < squareFadeRate * 7){
            squareFade[i + 2 + j * realWide] += 8;
          } 
          else {
            squareFade[i + 2 + j * realWide] = (squareFadeRate * 8) - 1;
          }
        }
      }
    }
    if(drawBorder){
      for(int i = 0; i < realWide; i++){
        for(int j = 0; j < high + 2; j++){
          if(square[i + j * realWide] > 0){
            int xDraw = x + (i - 2) * sizex;
            int yDraw = y + j * sizey;
            rect(xDraw, yDraw, sizex, sizey);
          }
        }
      }
    }
    //timed drop of block in field
    if(millis() - timer > dropRate){
      block.drop(false);
      timer = millis();
    }
  }
  //function to check if a row in the array is full
  boolean fullRow(int row){
    int trueSquares = 0;
    for(int i = 0; i < wide; i++){
      if(square[i + 2 + row * realWide] > 0){
        trueSquares++;
      }
    }
    if(trueSquares == wide){
      lines++;
      println("lines:"+lines);
      //speed increase by factoring equation
      if(float(lines) / increaseSpeed == lines / increaseSpeed){
        dropRate = int(float(dropRate) * 0.8);
      }
      int [] temp = new int[wide];
      arraycopy(square, 2 + row * realWide, temp, 0, wide); 
      building.addRow(temp);
      return true;
    }
    return false;
  }
  //function to look for full rows in the TetrisField
  void checkRows(){
    int countLines = 0;
    for(int i = 0; i < high; i++){
      if(fullRow(i)){
        for(int j = 0; j < wide; j++){
          square[j + 2 + i * realWide] = 0;
          squareFade[j + 2 + i * realWide] = 0;
        }
        for(int j = i; j > 0; j--){
          for(int k = 0; k < wide; k++){
            square[k + 2 + j * realWide] = square[k + 2 + (j - 1) * realWide];
            squareFade[k + 2 + j * realWide] = squareFade[k + 2 + (j - 1) * realWide];
          }
        }
        countLines++;
      }
    }
    //insert score increase based on countLines
    if(countLines > 0){
      score += 10 * sq(countLines);
      for(int i = ranking; i < rankTable.length; i++){     
        if(score >= rankTable[i] && i > ranking){
          ranking = i;
          rankingMugShot = int(random(mugShotTable[i]));
        }
      }
    }
  }
}

//blocks own a 4*4 or 3*3 space of squares, the block proper is a series of positive values in that space.
//the block's id sets its solid elements, dimensions for rotation and colour
//there is a collision check in the block's constructor to detect for the game over clause
//the constructor reads its values from TetrisNextBlock - that must be initialised first
class TetrisBlock{
  int [] square;
  int id, x, y, depth, wide, high;
  boolean initNextBlock = true;
  boolean checkRows = false;
  TetrisBlock(){
    x = -1 + field.wide>>1;                         //handling point of block
    y = 0;                                          //handling point of block
    depth = blockDepth;                             //isometric depth of block
    this.id = nextBlock.id;                         //block type and initial colour value
    square = new int[nextBlock.square.length];      //block array elements
    wide = nextBlock.wide;                          //width of block array
    high = nextBlock.high;                          //height of block array
    arraycopy(nextBlock.square, square);            //square elements of block copied from nextBlock
    //game over collision detection
    for(int i = 0; i < wide; i++){
      for(int j = 0; j < high; j++){
        if(square[i + j * wide] > 0 && field.square[(x + 2 + i) + (y + j) * field.realWide] > 0){
          if(field.score > bestScore){
            bestScore = field.score;
          }
          println("game over");
          gameState = gameOver;
          lastScore = field.score;
          lastLines = field.lines;
          lastRanking = ranking;
          ranking = 0;
          rankingMugShot = int(random(mugShotTable[0]));
          quadTextSize(gameOverSize);
          String s0 = "you scored " + lastScore + " points";
          String s1 = "and made " + lastLines + " line";
          if(lastLines != 1){
            s1 += "s";
          }
          String s2 = "your rank is " + rankName[lastRanking];
          youScoredText = new QuadTextWobble(s0, (width / 2) - (quadTextWidth(s0) / 2), (height / 2) - 80, wobbleRange, wobbleSpeed);
          andMadeText = new QuadTextWobble(s1, (width / 2) - (quadTextWidth(s1) / 2), (height / 2), wobbleRange, wobbleSpeed);
          yourRankText = new QuadTextWobble(s2, (width / 2) - (quadTextWidth(s2) / 2), (height / 2) + 80, wobbleRange, wobbleSpeed);
          field = new TetrisField(field.x, field.y, field.wide, field.high, field.sizex, field.sizey, dropRate, field.increaseSpeed);
        }
      }
    }
  }
  void draw(){
    for(int i = 0; i < wide; i++){
      for(int j = 0; j < high; j++){
        if(square[i + j * wide] > 0){
          int xDraw = field.x + ((x + i) * field.sizex);
          int yDraw = field.y + ((y + j) * field.sizey);
          architecture[square[i + j * wide]].updatePixels();
          stroke(255);
          fill(blockColorEdge[id]);
          beginShape(QUADS);
          texture(architecture[square[i + j * wide]]);
          vertex(xDraw - (depth>>1), yDraw - (depth>>1), -depth, 0, 0);
          vertex(xDraw, yDraw, 0, field.sizex, 0);
          vertex(xDraw, yDraw + field.sizey, 0, field.sizex, field.sizey);
          vertex(xDraw - (depth>>1), yDraw + field.sizey - (depth>>1), -depth, 0, field.sizey);
          endShape();
          fill(blockColorEdge[id]);
          beginShape(QUADS);
          texture(architecture[square[i + j * wide]]);
          vertex(xDraw - (depth>>1), yDraw - (depth>>1), -depth, 0, 0);
          vertex(xDraw + field.sizex - (depth>>1), yDraw - (depth>>1), -depth, field.sizex, 0);
          vertex(xDraw + field.sizex, yDraw, 0, field.sizex, field.sizey);
          vertex(xDraw, yDraw, 0, 0, field.sizey);
          endShape();
          fill(blockColor[id+64]);
          beginShape(QUADS);
          texture(architecture[square[i + j * wide]]);
          vertex(xDraw, yDraw, 0, 0, 0);
          vertex(xDraw + field.sizex, yDraw, 0, field.sizex, 0);
          vertex(xDraw + field.sizex, yDraw + field.sizey, 0, field.sizex, field.sizey);
          vertex(xDraw, yDraw + field.sizey, 0, 0, field.sizey);
          endShape();
        }
      }
    }
  }
  //spin contents of block matrix clockwise
  void rotate(){
    int [] temp = new int[square.length];
    boolean canRotate = true;
    for(int i = 0; i < wide; i++){
      int start = (wide - 1 - i) * wide;
      for(int j = 0; j < high; j++){
        temp[i + j * wide] = square[start + j];
        if(temp[i + j * wide] > 0 && field.square[(x + 2 + i) + (y + j) * field.realWide] > 0){
          canRotate = false;
          break;
        }
      }
      if(!canRotate){
        break;
      }
    }
    if(canRotate){
      for(int i = 0; i < square.length; i++){
        square[i] = temp[i];
      }
    }
  }
  //collision detection against TetrisField and scroll down block, if the block is not manually dropped
  //darken the colour of the block by increasing its colour id
  void drop(boolean manual){
    for(int i = 0; i < wide; i++){
      for(int j= 0; j < high; j++){
        if(square[i + j * wide] > 0 && field.square[(x + 2 + i) + (y + j + 1) * field.realWide] > 0){
          convert();
        }
      }
    }
    y++;
  }
  //move block according to direction value and check for collisions along block's edge
  void move(int direction){
    boolean freeToMove = true;
    for(int i = 0; i < wide; i++){
      for(int j = 0; j < high; j++){
        if(square[i + j * wide] > 0 && field.square[(x + 2 + i + direction) + (y + j) * field.realWide] > 0){
          freeToMove = false;
        }
      }
    }
    if(freeToMove){
      x += direction;
    }
  }
  //turn block into TetrisField elements
  void convert(){
    for(int i = 0; i < wide; i++){
      for(int j = 0; j < high; j++){
        if(square[i + j * wide] > 0){
          field.square[(x + 2 + i) + (y + j) * field.realWide] = square[i + j * wide];
          field.squareFade[(x + 2 + i) + (y + j) * field.realWide] = id;
        }
      }
    }
    checkRows = true;
  }
}

class TetrisNextBlock{
  int [] square;
  int id, x, y, depth, wide, high;
  //below are the different types of block arrays
  int [][] blockType = {
    {
      0, 0, 0, 0,
      0, 1, 1, 0,
      0, 1, 1, 0,
      0, 0, 0, 0                                                            }
    ,{
      0, 1, 0, 0,
      0, 1, 0, 0,
      0, 1, 0, 0,
      0, 1, 0, 0                                                            }
    ,{
      0, 0, 0,
      1, 1, 1,
      0, 1, 0                                                            }
    ,{
      0, 1, 0,
      0, 1, 0,
      1, 1, 0                                                            }
    ,{
      0, 1, 0,
      0, 1, 0,
      0, 1, 1                                                            }
    ,{
      1, 0, 0,
      1, 1, 0,
      0, 1, 0                                                            }
    ,{
      0, 1, 0,
      1, 1, 0,
      1, 0, 0                                                            }
  };
  TetrisNextBlock(){
    x = -1 + field.wide>>1;                         //handling point of block
    y = 0;                                          //handling point of block
    depth = blockDepth;                             //isometric depth of block
    this.id = int(random(7) + 1);                   //block type and initial colour value
    square = new int[16];                           //block array elements
    wide = 4;                                       //width of block array
    high = 4;                                       //height of block array
    //resizing of array for more compact blocks
    if(id > 2){
      square = new int[9];
      wide = 3;
      high = 3;
    }
    //colour setting
    arraycopy(blockType[id - 1], square);
    for(int i = 0; i < square.length; i++){
      if(random(1) > 0.7){
        square[i] *= int(random(imagesToLoad - 1)) + 2;
      }
    }
  }
  void draw(int xPlace, int yPlace){
    for(int i = 0; i < wide; i++){
      for(int j = 0; j < high; j++){
        if(square[i + j * wide] > 0){
          int xDraw = xPlace + ((x + i) * field.sizex);
          int yDraw = yPlace + ((y + j) * field.sizey);
          architecture[square[i + j * wide]].updatePixels();
          fill(blockColorEdge[id]);
          stroke(255);
          beginShape(QUADS);
          texture(architecture[square[i + j * wide]]);
          vertex(xDraw - (depth>>1), yDraw - (depth>>1), -depth, 0, 0);
          vertex(xDraw, yDraw, 0, field.sizex, 0);
          vertex(xDraw, yDraw + field.sizey, 0, field.sizex, field.sizey);
          vertex(xDraw - (depth>>1), yDraw + field.sizey - (depth>>1), -depth, 0, field.sizey);
          endShape();
          fill(blockColorEdge[id]);
          beginShape(QUADS);
          texture(architecture[square[i + j * wide]]);
          vertex(xDraw - (depth>>1), yDraw - (depth>>1), -depth, 0, 0);
          vertex(xDraw + field.sizex - (depth>>1), yDraw - (depth>>1), -depth, field.sizex, 0);
          vertex(xDraw + field.sizex, yDraw, 0, field.sizex, field.sizey);
          vertex(xDraw, yDraw, 0, 0, field.sizey);
          endShape();
          fill(blockColor[id+64]);
          beginShape(QUADS);
          texture(architecture[square[i + j * wide]]);
          vertex(xDraw, yDraw, 0, 0, 0);
          vertex(xDraw + field.sizex, yDraw, 0, field.sizex, 0);
          vertex(xDraw + field.sizex, yDraw + field.sizey, 0, field.sizex, field.sizey);
          vertex(xDraw, yDraw + field.sizey, 0, 0, field.sizey);
          endShape();
        }
      }
    }
  }
}

class Building{
  int [] square;
  int wide, high, sizex, sizey;
  Building(){
    wide = field.wide;
    high = 0;
    sizex = field.sizex >> 1;
    sizey = field.sizey >> 1;
  }
  void draw(int x, int y){
    noStroke();
    if(square != null){
      for(int i = 0; i < wide; i++){
        for(int j = 0; j < high; j++){
          if(square[i + j * wide] > 0){
            int xDraw = x + i * sizex;
            int yDraw = y + j * sizey;
            int depth = 30;
            fill(200);
            architecture[square[i + j * wide]].updatePixels();
            fill(255);
            beginShape(QUADS);
            texture(architecture[square[i + j * wide]]);
            vertex(xDraw, yDraw, 0, 0, 0);
            vertex(xDraw + sizex, yDraw, 0, field.sizex, 0);
            vertex(xDraw + sizex, yDraw + sizey, 0, field.sizex, field.sizey);
            vertex(xDraw, yDraw + sizey, 0, 0, field.sizey);
            endShape();
          }
        }
      }
    }
  }
  //function for adding row to building
  void addRow(int [] row){
    if(square == null){
      square = new int[row.length];
      arraycopy(row, square);
      high = 1;
    }
    else{
      square = concat(row, square);
      high++;
    }
  }
}
